# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2020, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Converter to convert a problem with equality constraints to unconstrained with penalty terms."""

import logging
from typing import Optional, cast, Union, Tuple, List

import numpy as np

from .quadratic_program_converter import QuadraticProgramConverter
from ..exceptions import QiskitOptimizationError
from ..problems.constraint import Constraint
from ..problems.quadratic_objective import QuadraticObjective
from ..problems.quadratic_program import QuadraticProgram
from ..problems.variable import Variable

logger = logging.getLogger(__name__)


class LinearEqualityToPenalty(QuadraticProgramConverter):
    """Convert a problem with only equality constraints to unconstrained with penalty terms."""

    def __init__(self, penalty: Optional[float] = None) -> None:
        """
        Args:
            penalty: Penalty factor to scale equality constraints that are added to objective.
                     If None is passed, a penalty factor will be automatically calculated on
                     every conversion.
        """
        self._src_num_vars: Optional[int] = None
        self._penalty: Optional[float] = penalty
        self._should_define_penalty: bool = penalty is None

    def convert(self, problem: QuadraticProgram) -> QuadraticProgram:
        """Convert a problem with equality constraints into an unconstrained problem.

        Args:
            problem: The problem to be solved, that does not contain inequality constraints.

        Returns:
            The converted problem, that is an unconstrained problem.

        Raises:
            QiskitOptimizationError: If an inequality constraint exists.
        """

        # create empty QuadraticProgram model
        self._src_num_vars = problem.get_num_vars()
        dst = QuadraticProgram(name=problem.name)

        # If no penalty was given, set the penalty coefficient by _auto_define_penalty()
        if self._should_define_penalty:
            penalty = self._auto_define_penalty(problem)
        else:
            penalty = self._penalty

        # Set variables
        for x in problem.variables:
            if x.vartype == Variable.Type.CONTINUOUS:
                dst.continuous_var(x.lowerbound, x.upperbound, x.name)
            elif x.vartype == Variable.Type.BINARY:
                dst.binary_var(x.name)
            elif x.vartype == Variable.Type.INTEGER:
                dst.integer_var(x.lowerbound, x.upperbound, x.name)
            else:
                raise QiskitOptimizationError(f"Unsupported vartype: {x.vartype}")

        # get original objective terms
        offset = problem.objective.constant
        linear = problem.objective.linear.to_dict()
        quadratic = problem.objective.quadratic.to_dict()
        sense = problem.objective.sense.value

        # convert linear constraints into penalty terms
        for constraint in problem.linear_constraints:
            if constraint.sense != Constraint.Sense.EQ:
                raise QiskitOptimizationError(
                    "An inequality constraint exists. "
                    "The method supports only equality constraints."
                )

            constant = constraint.rhs
            row = constraint.linear.to_dict()

            # constant parts of penalty*(Constant-func)**2: penalty*(Constant**2)
            offset += sense * penalty * constant**2

            # linear parts of penalty*(Constant-func)**2: penalty*(-2*Constant*func)
            for j, coef in row.items():
                # if j already exists in the linear terms dic, add a penalty term
                # into existing value else create new key and value in the linear_term dict
                linear[j] = linear.get(j, 0.0) + sense * penalty * -2 * coef * constant

            # quadratic parts of penalty*(Constant-func)**2: penalty*(func**2)
            for j, coef_1 in row.items():
                for k, coef_2 in row.items():
                    # if j and k already exist in the quadratic terms dict,
                    # add a penalty term into existing value
                    # else create new key and value in the quadratic term dict

                    # according to implementation of quadratic terms in OptimizationModel,
                    # don't need to multiply by 2, since loops run over (x, y) and (y, x).
                    tup = cast(Union[Tuple[int, int], Tuple[str, str]], (j, k))
                    quadratic[tup] = (
                        quadratic.get(tup, 0.0) + sense * penalty * coef_1 * coef_2
                    )

        if problem.objective.sense == QuadraticObjective.Sense.MINIMIZE:
            dst.minimize(offset, linear, quadratic)
        else:
            dst.maximize(offset, linear, quadratic)

        # Update the penalty to the one just used
        self._penalty = penalty

        return dst

    @staticmethod
    def _auto_define_penalty(problem: QuadraticProgram) -> float:
        """Automatically define the penalty coefficient.

        Returns:
            Return the minimum valid penalty factor calculated
            from the upper bound and the lower bound of the objective function.
            If a constraint has a float coefficient,
            return the default value for the penalty factor.
        """
        default_penalty = 1e5

        # Check coefficients of constraints.
        # If a constraint has a float coefficient, return the default value for the penalty factor.
        terms = []
        for constraint in problem.linear_constraints:
            terms.append(constraint.rhs)
            terms.extend(constraint.linear.to_array().tolist())
        if any(isinstance(term, float) and not term.is_integer() for term in terms):
            logger.warning(
                "Warning: Using %f for the penalty coefficient because "
                "a float coefficient exists in constraints. \n"
                "The value could be too small. "
                "If so, set the penalty coefficient manually.",
                default_penalty,
            )
            return default_penalty

        lin_b = problem.objective.linear.bounds
        quad_b = problem.objective.quadratic.bounds
        return (
            1.0
            + (lin_b.upperbound - lin_b.lowerbound)
            + (quad_b.upperbound - quad_b.lowerbound)
        )

    def interpret(self, x: Union[np.ndarray, List[float]]) -> np.ndarray:
        """Convert the result of the converted problem back to that of the original problem

        Args:
            x: The result of the converted problem or the given result in case of FAILURE.

        Returns:
            The result of the original problem.

        Raises:
            QiskitOptimizationError: if the number of variables in the result differs from
                                     that of the original problem.
        """
        if len(x) != self._src_num_vars:
            raise QiskitOptimizationError(
                "The number of variables in the passed result differs from "
                "that of the original problem."
            )
        return np.asarray(x)

    @property
    def penalty(self) -> Optional[float]:
        """Returns the penalty factor used in conversion.

        Returns:
            The penalty factor used in conversion.
        """
        return self._penalty

    @penalty.setter
    def penalty(self, penalty: Optional[float]) -> None:
        """Set a new penalty factor.

        Args:
            penalty: The new penalty factor.
                     If None is passed, a penalty factor will be automatically calculated
                     on every conversion.
        """
        self._penalty = penalty
        self._should_define_penalty = penalty is None



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Optimization converters (:mod:`qiskit_optimization.converters`)
===============================================================

.. currentmodule:: qiskit_optimization.converters

This is a set of converters having `convert` functionality to go between different representations
of a given :class:`~qiskit_optimization.problems.QuadraticProgram` and to `interpret` a given
result for the problem, based on the original problem before conversion, to return an appropriate
:class:`~qiskit_optimization.algorithms.OptimizationResult`.

Base class for converters
-------------------------

.. autosummary::
   :toctree: ../stubs/
   :nosignatures:

   QuadraticProgramConverter

Converters
----------

.. autosummary::
   :toctree: ../stubs/
   :nosignatures:

   InequalityToEquality
   IntegerToBinary
   LinearEqualityToPenalty
   LinearInequalityToPenalty
   MaximizeToMinimize
   MinimizeToMaximize
   QuadraticProgramToQubo

"""

from .integer_to_binary import IntegerToBinary
from .inequality_to_equality import InequalityToEquality
from .linear_equality_to_penalty import LinearEqualityToPenalty
from .linear_inequality_to_penalty import LinearInequalityToPenalty
from .flip_problem_sense import MaximizeToMinimize
from .flip_problem_sense import MinimizeToMaximize
from .quadratic_program_to_qubo import QuadraticProgramToQubo
from .quadratic_program_converter import QuadraticProgramConverter

__all__ = [
    "InequalityToEquality",
    "IntegerToBinary",
    "LinearEqualityToPenalty",
    "LinearInequalityToPenalty",
    "MaximizeToMinimize",
    "MinimizeToMaximize",
    "QuadraticProgramConverter",
    "QuadraticProgramToQubo",
]



from __future__ import annotations

from pydantic import BaseModel

from q3as.algo.optimizer import Optimizer


class EncodedOptimizer(BaseModel, Optimizer):
    method: str

    @classmethod
    def encode(cls, optimizer: Optimizer) -> EncodedOptimizer:
        return cls(method=optimizer.scipy_method)

    @property
    def scipy_method(self) -> str:
        return self.method



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2020, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Converter to convert a problem with inequality constraints to unconstrained with penalty terms."""

import logging
from typing import Optional, Union, Tuple, List, Dict

import numpy as np

from .quadratic_program_converter import QuadraticProgramConverter
from ..exceptions import QiskitOptimizationError
from ..problems.constraint import Constraint, ConstraintSense
from ..problems.quadratic_objective import QuadraticObjective
from ..problems.quadratic_program import QuadraticProgram
from ..problems.variable import Variable

logger = logging.getLogger(__name__)


class LinearInequalityToPenalty(QuadraticProgramConverter):
    r"""Convert linear inequality constraints to penalty terms of the objective function.

    There are some linear constraints which do not require slack variables to
    construct penalty terms [1]. This class supports the following inequality constraints.

    .. math::

        \begin{array}{}
        \text { Inequality constraint } & & \text { Penalty term } \\
        x \leq y & \rightarrow  & P(x-x y) \\
        x \geq y & \rightarrow  & P(y-x y) \\
        \sum_{i=1}^n x_i \leq 1, n \geq 2 & \rightarrow & P \sum_{i, j : i < j} x_i x_j\\
        \sum_{i=1}^n x_i \geq n-1, n \geq 2 & \rightarrow & P \sum_{i, j : i < j} (1 - x_i) (1 - x_j)
        \end{array}

    Note that x, y, z and :math:`x_i` are binary variables, and P is a penalty factor,
    where the value of P is automatically determined or supplied by users.

    If constraints match with any of the patterns, they are converted into penalty terms and added
    to the objective function. Otherwise, constraints are kept as is.

    References:
        [1]: Fred Glover, et al. (2019),
             A Tutorial on Formulating and Using QUBO Models,
             `arXiv:1811.11538 <https://arxiv.org/abs/1811.11538>`_.

    """

    def __init__(self, penalty: Optional[float] = None) -> None:
        """
        Args:
            penalty: Penalty factor to scale equality constraints that are added to objective.
                     If None is passed, a penalty factor will be automatically calculated on
                     every conversion.
        """

        self._src_num_vars: Optional[int] = None
        self._dst: Optional[QuadraticProgram] = None
        self._penalty: Optional[float] = penalty
        self._should_define_penalty: bool = penalty is None

    def convert(self, problem: QuadraticProgram) -> QuadraticProgram:
        r"""Convert inequality constraints into penalty terms of the objective function.

        This methods converts the following patterns where x, y, and :math:`x_i` are binary variables
        and P is a penalty factor.

        .. math::

            \begin{array}{}
            \text { Inequality constraint } & & \text { Penalty term } \\
            x \leq y & \rightarrow  & P(x-x y) \\
            x \geq y & \rightarrow  & P(y-x y) \\
            \sum_{i=1}^n x_i \leq 1, n \geq 2 & \rightarrow & P \sum_{i, j : i < j} x_i x_j\\
            \sum_{i=1}^n x_i \geq n-1, n \geq 2 & \rightarrow & P \sum_{i, j : i < j} (1 - x_i) (1 - x_j)
            \end{array}

        Args:
            problem: The problem to be solved.

        Returns:
            The converted problem

        Raises:
            QiskitOptimizationError: If an unsupported-type variable exists.
        """

        # create empty QuadraticProgram model
        self._src_num_vars = problem.get_num_vars()
        self._dst = QuadraticProgram(name=problem.name)

        # If no penalty was given, set the penalty coefficient by _auto_define_penalty()
        if self._should_define_penalty:
            penalty = self._auto_define_penalty(problem)
        else:
            penalty = self._penalty

        # Set variables
        for x in problem.variables:
            if x.vartype == Variable.Type.CONTINUOUS:
                self._dst.continuous_var(x.lowerbound, x.upperbound, x.name)
            elif x.vartype == Variable.Type.BINARY:
                self._dst.binary_var(x.name)
            elif x.vartype == Variable.Type.INTEGER:
                self._dst.integer_var(x.lowerbound, x.upperbound, x.name)
            else:
                raise QiskitOptimizationError(f"Unsupported vartype: {x.vartype}")

        # get original objective terms
        offset = problem.objective.constant
        linear = problem.objective.linear.to_dict()
        quadratic = problem.objective.quadratic.to_dict()
        sense = problem.objective.sense.value

        # convert linear constraints into penalty terms
        for constraint in problem.linear_constraints:
            # special constraint check function here
            if not self._is_matched_constraint(problem, constraint):
                self._dst.linear_constraint(
                    constraint.linear.coefficients,
                    constraint.sense,
                    constraint.rhs,
                    constraint.name,
                )
                continue

            conv_offset, conv_linear, conv_quadratic, varmap = self._conversion_table(
                constraint
            )

            # constant part
            offset += sense * penalty * conv_offset

            # linear parts of penalty
            for j, j_2 in varmap.items():
                # if j already exists in the linear terms dic, add a penalty term
                # into existing value else create new key and value in the linear_term dict

                if conv_linear[j] != 0:
                    linear[j_2] = (
                        linear.get(j_2, 0.0) + sense * penalty * conv_linear[j]
                    )

            # quadratic parts of penalty
            for j, j_2 in varmap.items():
                for k in range(j, len(varmap)):
                    # if j and k already exist in the quadratic terms dict,
                    # add a penalty term into existing value
                    # else create new key and value in the quadratic term dict

                    if conv_quadratic[j][k] != 0:
                        tup = (j_2, varmap[k])
                        quadratic[tup] = (
                            quadratic.get(tup, 0.0)
                            + sense * penalty * conv_quadratic[j][k]
                        )

        # Copy quadratic_constraints
        for quadratic_constraint in problem.quadratic_constraints:
            self._dst.quadratic_constraint(
                quadratic_constraint.linear.coefficients,
                quadratic_constraint.quadratic.coefficients,
                quadratic_constraint.sense,
                quadratic_constraint.rhs,
                quadratic_constraint.name,
            )

        if problem.objective.sense == QuadraticObjective.Sense.MINIMIZE:
            self._dst.minimize(offset, linear, quadratic)
        else:
            self._dst.maximize(offset, linear, quadratic)

        # Update the penalty to the one just used
        self._penalty = penalty

        return self._dst

    @staticmethod
    def _conversion_table(
        constraint,
    ) -> Tuple[int, np.ndarray, np.ndarray, Dict[int, int]]:
        """Construct conversion matrix for special constraint.

        Returns:
            Return conversion table which is used to construct
            penalty term in main function.

        Raises:
            QiskitOptimizationError: if the constraint is invalid.
        """

        vars_dict = constraint.linear.to_dict()
        coeffs = list(vars_dict.values())
        varmap = dict(enumerate(vars_dict.keys()))
        rhs = constraint.rhs
        sense = constraint.sense

        num_vars = len(vars_dict)

        # initialize return values, these are used for converted offset, linear
        # and quadratic terms
        offset = 0
        linear = np.zeros(num_vars, dtype=int)
        quadratic = np.zeros((num_vars, num_vars), dtype=int)

        # rhs = num_vars - 1 correspond to multiple variable with >= n - 1 case.
        if sense == ConstraintSense.GE and rhs == num_vars - 1:
            # x_1 + ... + x_n >= n - 1
            # The number of offset is combination ( nC2 )
            offset = num_vars * (num_vars - 1) // 2
            linear = np.full(num_vars, 1 - num_vars, dtype=int)
            quadratic = np.triu(np.ones((num_vars, num_vars), dtype=int), k=1)
        elif sense == ConstraintSense.LE and rhs == 1:
            # x_1 + ... + x_n <= 1
            quadratic = np.triu(np.ones((num_vars, num_vars), dtype=int), k=1)
        elif rhs == 0:
            if num_vars != 2:
                raise QiskitOptimizationError(
                    f"Internal error: invalid number of variables {num_vars} {constraint.name}"
                )
            quadratic = np.array([[0, -1], [0, 0]])
            if sense == ConstraintSense.GE:
                # x >= y case
                if coeffs[0] < 0.0:
                    linear[0] = 1
                else:
                    linear[1] = 1
            elif sense == ConstraintSense.LE:
                # x <= y case
                if coeffs[0] > 0.0:
                    linear[0] = 1
                else:
                    linear[1] = 1
        else:
            raise QiskitOptimizationError(
                f"Internal error: invalid constraint {constraint.name}"
            )

        return offset, linear, quadratic, varmap

    @staticmethod
    def _is_matched_constraint(problem, constraint) -> bool:
        """Determine if constraint is special or not.

        Returns:
            True: when constraint is special
            False: when constraint is not special
        """

        params = constraint.linear.to_dict()
        num_vars = len(params)
        rhs = constraint.rhs
        sense = constraint.sense
        coeff_array = np.array(list(params.values()))

        # Binary parameter?
        if any(
            problem.variables[i].vartype != Variable.Type.BINARY for i in params.keys()
        ):
            return False

        if num_vars == 2 and rhs == 0:
            if sense in (Constraint.Sense.LE, Constraint.Sense.GE):
                # x-y<=0
                # x-y>=0
                return coeff_array.min() == -1.0 and coeff_array.max() == 1.0
        elif num_vars >= 2:
            if sense == Constraint.Sense.LE and rhs == 1:
                if all(i == 1 for i in params.values()):
                    # x1+x2+...<=1
                    return True
            elif sense == Constraint.Sense.GE and rhs == num_vars - 1:
                if all(i == 1 for i in params.values()):
                    # x1+x2+...>=n-1
                    return True

        return False

    @staticmethod
    def _auto_define_penalty(problem) -> float:
        """Automatically define the penalty coefficient.

        Returns:
            Return the minimum valid penalty factor calculated
            from the upper bound and the lower bound of the objective function.
            If a constraint has a float coefficient,
            return the default value for the penalty factor.
        """

        default_penalty = 1e5

        # Check coefficients of constraints.
        # If a constraint has a float coefficient, return the default value for the penalty factor.
        terms = []
        for constraint in problem.linear_constraints:
            terms.append(constraint.rhs)
            terms.extend(constraint.linear.to_array().tolist())
        if any(isinstance(term, float) and not term.is_integer() for term in terms):
            logger.warning(
                "Warning: Using %f for the penalty coefficient because "
                "a float coefficient exists in constraints. \n"
                "The value could be too small. "
                "If so, set the penalty coefficient manually.",
                default_penalty,
            )
            return default_penalty

        lin_b = problem.objective.linear.bounds
        quad_b = problem.objective.quadratic.bounds
        return (
            1.0
            + (lin_b.upperbound - lin_b.lowerbound)
            + (quad_b.upperbound - quad_b.lowerbound)
        )

    def interpret(self, x: Union[np.ndarray, List[float]]) -> np.ndarray:
        """Convert the result of the converted problem back to that of the original problem

        Args:
            x: The result of the converted problem or the given result in case of FAILURE.

        Returns:
            The result of the original problem.

        Raises:
            QiskitOptimizationError: if the number of variables in the result differs from
                                     that of the original problem.
        """

        if len(x) != self._src_num_vars:
            raise QiskitOptimizationError(
                f"The number of variables in the passed result ({len(x)}) differs from "
                f"that of the original problem ({self._src_num_vars})."
            )
        return np.asarray(x)

    @property
    def penalty(self) -> Optional[float]:
        """Returns the penalty factor used in conversion.

        Returns:
            The penalty factor used in conversion.
        """

        return self._penalty

    @penalty.setter
    def penalty(self, penalty: Optional[float]) -> None:
        """Set a new penalty factor.

        Args:
            penalty: The new penalty factor.
                     If None is passed, a penalty factor will be automatically calculated
                     on every conversion.
        """

        self._penalty = penalty
        self._should_define_penalty = penalty is None



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Abstract Constraint."""

from abc import abstractmethod
from enum import Enum
from typing import Union, List, Dict, Any

from numpy import ndarray

from .quadratic_program_element import QuadraticProgramElement
from ..exceptions import QiskitOptimizationError


class ConstraintSense(Enum):
    """Constraint Sense Type."""

    # pylint: disable=invalid-name
    LE = 0
    GE = 1
    EQ = 2

    @staticmethod
    def convert(sense: Union[str, "ConstraintSense"]) -> "ConstraintSense":
        """Convert a string into a corresponding sense of constraints

        Args:
            sense: A string or sense of constraints

        Returns:
            The sense of constraints

        Raises:
            QiskitOptimizationError: if the input string is invalid.
        """
        if isinstance(sense, ConstraintSense):
            return sense
        sense = sense.upper()
        if sense not in [
            "E",
            "L",
            "G",
            "EQ",
            "LE",
            "GE",
            "=",
            "==",
            "<=",
            "<",
            ">=",
            ">",
        ]:
            raise QiskitOptimizationError(f"Invalid sense: {sense}")
        if sense in ["E", "EQ", "=", "=="]:
            return ConstraintSense.EQ
        elif sense in ["L", "LE", "<=", "<"]:
            return ConstraintSense.LE
        else:
            return ConstraintSense.GE

    @property
    def label(self) -> str:
        """Label of the constraint sense

        Returns:
            The label of the constraint sense ('<=', '>=', or '==')
        """
        if self is ConstraintSense.LE:
            return "<="
        elif self is ConstraintSense.GE:
            return ">="
        else:
            return "=="


class Constraint(QuadraticProgramElement):
    """Abstract Constraint Class."""

    Sense = ConstraintSense

    def __init__(
        self, quadratic_program: Any, name: str, sense: ConstraintSense, rhs: float
    ) -> None:
        """Initializes the constraint.

        Args:
            quadratic_program: The parent QuadraticProgram.
            name: The name of the constraint.
            sense: The sense of the constraint.
            rhs: The right-hand-side of the constraint.
        """
        super().__init__(quadratic_program)
        self._name = name
        self._sense = sense
        self._rhs = rhs

    @property
    def name(self) -> str:
        """Returns the name of the constraint.

        Returns:
            The name of the constraint.
        """
        return self._name

    @property
    def sense(self) -> ConstraintSense:
        """Returns the sense of the constraint.

        Returns:
            The sense of the constraint.
        """
        return self._sense

    @sense.setter
    def sense(self, sense: ConstraintSense) -> None:
        """Sets the sense of the constraint.

        Args:
            sense: The sense of the constraint.
        """
        self._sense = sense

    @property
    def rhs(self) -> float:
        """Returns the right-hand-side of the constraint.

        Returns:
            The right-hand-side of the constraint.
        """
        return self._rhs

    @rhs.setter
    def rhs(self, rhs: float) -> None:
        """Sets the right-hand-side of the constraint.

        Args:
            rhs: The right-hand-side of the constraint.
        """
        self._rhs = rhs

    @abstractmethod
    def evaluate(self, x: Union[ndarray, List, Dict[Union[int, str], float]]) -> float:
        """Evaluate left-hand-side of constraint for given values of variables.

        Args:
            x: The values to be used for the variables.

        Returns:
            The left-hand-side of the constraint.
        """
        raise NotImplementedError()



from __future__ import annotations
from typing import Literal

import io
import base64
import numpy as np
from numpy.typing import ArrayLike
from pydantic import BaseModel


class EncodedArray(BaseModel):
    encoding: Literal["npy"] = "npy"
    base64: bytes

    @classmethod
    def encode(cls, array: ArrayLike) -> EncodedArray:
        data = io.BytesIO()
        np.save(data, array, allow_pickle=False)
        return cls(encoding="npy", base64=base64.b64encode(data.getvalue()))

    def decode(self) -> np.ndarray:
        data = base64.b64decode(self.base64)
        return np.load(io.BytesIO(data), allow_pickle=False)



from __future__ import annotations
from typing import Any

from pydantic import BaseModel

from q3as.app.application import Application, ApplicationName

import q3as.app.qubo
import q3as.app.maxcut


class EncodedApplication(BaseModel):
    name: ApplicationName
    data: Any

    @classmethod
    def encode(cls, app: Application) -> EncodedApplication:
        return cls(name=app.name(), data=app.encode())

    def decode(self) -> Application:
        if self.name == "qubo":
            return q3as.app.qubo.Qubo.decode_any(self.data)
        if self.name == "maxcut":
            return q3as.app.maxcut.Maxcut.decode_any(self.data)
        raise ValueError(f"Unknown application name: {self.name}")



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2022, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Translate ``QuadraticProgram`` into a pretty-printed string"""

from io import StringIO
from math import isclose
from typing import List, Optional, Union, cast

import numpy as np

from ..infinity import INFINITY
from ..exceptions import QiskitOptimizationError
from ..problems import (
    LinearExpression,
    QuadraticExpression,
    QuadraticObjective,
    QuadraticProgram,
    VarType,
)

DEFAULT_TRUNCATE = 50


def _int_if_close(val: Union[int, float, np.integer, np.floating]) -> Union[int, float]:
    """Convert a value into an integer if possible

    Note: if abs(val) is too large, int(val) is not correct
          e.g., int(1e16 - 1) -> 10000000000000000
    """
    if isinstance(val, np.integer):
        val = int(val)
    elif isinstance(val, np.floating):
        val = float(val)

    if isinstance(val, int):
        return val
    if abs(val) <= 1e10 and val.is_integer():
        return int(val)
    return val


def _term2str(coeff: float, term: str, is_head: bool) -> str:
    """Translate a pair of a coefficient and a term to a string.

    Args:
        coeff: a coefficient.
        term: a term. This can be empty and `coeff` is treated as a constant.
        is_head: Whether this coefficient appears in the head of the string or not.

    Returns:
        A strings representing the term.
    """
    if term:
        if is_head:
            if isclose(coeff, 1.0):
                ret = term
            elif isclose(coeff, -1.0):
                ret = f"-{term}"
            else:
                ret = f"{_int_if_close(coeff)}*{term}"
        else:
            sign = "-" if coeff < 0.0 else "+"
            abs_val = abs(coeff)
            if isclose(abs_val, 1.0):
                ret = f"{sign} {term}"
            else:
                ret = f"{sign} {_int_if_close(abs_val)}*{term}"
    else:
        if is_head:
            ret = f"{_int_if_close(coeff)}"
        else:
            sign = "-" if coeff < 0.0 else "+"
            abs_val = abs(coeff)
            ret = f"{sign} {_int_if_close(abs_val)}"
    return ret


def _check_name(name: str, name_type: str) -> None:
    """Check a name is printable or not.

    Args:
        name: a variable name.
        name_type: the type associated with the name.

    Raises:
        QiskitOptimizationError: if the name is not printable.
    """
    if not name.isprintable():
        raise QiskitOptimizationError(
            f"{name_type} name is not printable: {repr(name)}"
        )


def _concatenate_terms(terms: List[str], wrap: int, indent: int) -> str:
    ind = " " * indent
    if wrap == 0:
        return ind + " ".join(terms)
    buf = ind
    cur = indent
    for term in terms:
        if cur + len(term) >= wrap:
            buf += "\n"
            buf += ind
            cur = indent
        if cur != indent:  # if the position is not the start of the line
            buf += " "
            cur += 1
        buf += term
        cur += len(term)
    return buf


def expr2str(
    constant: float = 0.0,
    linear: Optional[LinearExpression] = None,
    quadratic: Optional[QuadraticExpression] = None,
    truncate: int = 0,
    suffix: str = "",
    wrap: int = 0,
    indent: int = 0,
) -> str:
    """Translate a combination of a constant, a linear expression, and a quadratic expression
    into a string.

    Args:
        constant: a constant part.
        linear: a linear expression.
        quadratic: a quadratic expression.
        truncate: the threshold of the output string to be truncated. If a string is longer than
            the threshold, it is truncated and appended "...", e.g., "x^2 + y +...".
            It is disabled by setting 0. The default value is 0.
        suffix: a suffix text.
        wrap: The text width to wrap the output strings. It is disabled by setting 0.
            Note that some strings might exceed this value, for example, a long variable
            name won't be wrapped. The default value is 0.
        indent: The indent size. The default value is 0.

    Returns:
        A string representing the combination of the expressions.

    Raises:
        ValueError: if ``truncate`` is negative.
        QiskitOptimizationError: if the variable name is not printable.
    """
    if truncate < 0:
        raise ValueError(f"Invalid truncate value: {truncate}")

    terms = []
    is_head = True
    lin_dict = linear.to_dict(use_name=True) if linear else {}
    quad_dict = quadratic.to_dict(use_name=True) if quadratic else {}

    # quadratic expression
    for (var1, var2), coeff in sorted(quad_dict.items()):
        _check_name(cast(str, var1), "Variable")
        _check_name(cast(str, var2), "Variable")
        if var1 == var2:
            terms.append(_term2str(coeff, f"{var1}^2", is_head))
        else:
            terms.append(_term2str(coeff, f"{var1}*{var2}", is_head))
        is_head = False

    # linear expression
    for var, coeff in sorted(lin_dict.items()):
        _check_name(cast(str, var), "Variable")
        terms.append(_term2str(coeff, f"{var}", is_head))
        is_head = False

    # constant
    if not isclose(constant, 0.0, abs_tol=1e-10):
        terms.append(_term2str(constant, "", is_head))
    elif not lin_dict and not quad_dict:
        terms.append(_term2str(0, "", is_head))

    # suffix
    if suffix:
        terms.append(suffix)

    ret = _concatenate_terms(terms, wrap, indent)
    if 0 < truncate < len(ret):
        ret = ret[:truncate] + "..."
    return ret


def prettyprint(quadratic_program: QuadraticProgram, wrap: int = 80) -> str:
    """Translate a :class:`~qiskit_optimization.problems.QuadraticProgram` into a pretty-printed string.

    Args:
        quadratic_program: The optimization problem to be translated into a string.
        wrap: The text width to wrap the output strings. It is disabled by setting 0.
            Note that some strings might exceed this value, for example, a long variable
            name won't be wrapped. The default value is 80.

    Returns:
        A pretty-printed string representing the problem.

    Raises:
        QiskitOptimizationError: if there is a non-printable name.
    """

    with StringIO() as buf:
        _check_name(quadratic_program.name, "Problem")
        buf.write(f"Problem name: {quadratic_program.name}\n\n")
        if quadratic_program.objective.sense == QuadraticObjective.Sense.MINIMIZE:
            buf.write("Minimize\n")
        else:
            buf.write("Maximize\n")
        buf.write(
            expr2str(
                quadratic_program.objective.constant,
                quadratic_program.objective.linear,
                quadratic_program.objective.quadratic,
                wrap=wrap,
                indent=2,
            )
        )
        buf.write("\n\nSubject to")
        num_lin_csts = quadratic_program.get_num_linear_constraints()
        num_quad_csts = quadratic_program.get_num_quadratic_constraints()
        if num_lin_csts == 0 and num_quad_csts == 0:
            buf.write("\n  No constraints\n")
        if num_lin_csts > 0:
            buf.write(f"\n  Linear constraints ({num_lin_csts})\n")
            for cst in quadratic_program.linear_constraints:
                _check_name(cst.name, "Linear constraint")
                suffix = f"{cst.sense.label} {_int_if_close(cst.rhs)}  '{cst.name}'\n"
                buf.write(
                    expr2str(linear=cst.linear, suffix=suffix, wrap=wrap, indent=4)
                )
        if num_quad_csts > 0:
            buf.write(f"\n  Quadratic constraints ({num_quad_csts})\n")
            for cst2 in quadratic_program.quadratic_constraints:
                _check_name(cst2.name, "Quadratic constraint")
                suffix = (
                    f"{cst2.sense.label} {_int_if_close(cst2.rhs)}  '{cst2.name}'\n"
                )
                buf.write(
                    expr2str(
                        linear=cst2.linear,
                        quadratic=cst2.quadratic,
                        suffix=suffix,
                        wrap=wrap,
                        indent=4,
                    )
                )
        if quadratic_program.get_num_vars() == 0:
            buf.write("\n  No variables\n")
        bin_vars = []
        int_vars = []
        con_vars = []
        for var in quadratic_program.variables:
            if var.vartype is VarType.BINARY:
                _check_name(var.name, "Variable")
                bin_vars.append(var.name)
            elif var.vartype is VarType.INTEGER:
                int_vars.append(var)
            else:
                con_vars.append(var)
        if int_vars:
            buf.write(f"\n  Integer variables ({len(int_vars)})\n")
            for var in int_vars:
                buf.write("    ")
                if var.lowerbound > -INFINITY:
                    buf.write(f"{_int_if_close(var.lowerbound)} <= ")
                _check_name(var.name, "Variable")
                buf.write(var.name)
                if var.upperbound < INFINITY:
                    buf.write(f" <= {_int_if_close(var.upperbound)}")
                buf.write("\n")
        if con_vars:
            buf.write(f"\n  Continuous variables ({len(con_vars)})\n")
            for var in con_vars:
                buf.write("    ")
                if var.lowerbound > -INFINITY:
                    buf.write(f"{_int_if_close(var.lowerbound)} <= ")
                _check_name(var.name, "Variable")
                buf.write(var.name)
                if var.upperbound < INFINITY:
                    buf.write(f" <= {_int_if_close(var.upperbound)}")
                buf.write("\n")
        if bin_vars:
            buf.write(f"\n  Binary variables ({len(bin_vars)})\n")
            buf.write(_concatenate_terms(bin_vars, wrap=wrap, indent=4))
            buf.write("\n")
        ret = buf.getvalue()
    return ret



import os

import numpy as np
from qiskit.circuit.library import EfficientSU2
from qiskit.quantum_info import SparsePauliOp

from q3as.algo.vqe import VQEResult
from q3as.run_options import RunOptions, EstimatorOptions, SamplerOptions
from q3as import Client, Credentials, VQE

url = os.getenv("Q3AS_URL", "http://localhost:8080")


async def test_client_with_init_params():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    ansatz = ansatz = EfficientSU2(hamiltonian.num_qubits)
    x0 = 2 * np.pi * np.random.random(ansatz.num_parameters)
    with Client(Credentials.load(".credentials.json"), url=url) as client:
        _ = (
            VQE.builder()
            .ansatz(ansatz)
            .observables(hamiltonian)
            .initial_params(x0)
            .send(client)
        )


async def test_client_no_init_params():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    ansatz = ansatz = EfficientSU2(hamiltonian.num_qubits)
    with Client(Credentials.load(".credentials.json"), url=url) as client:
        _ = VQE.builder().ansatz(ansatz).observables(hamiltonian).send(client)


async def test_client_scaleway():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    ansatz = ansatz = EfficientSU2(hamiltonian.num_qubits)
    with Client(Credentials.load(".credentials.json"), url=url) as client:
        _ = (
            VQE.builder()
            .ansatz(ansatz)
            .observables(hamiltonian)
            .maxiter(10)
            .send(
                client,
                RunOptions(
                    backend="scaleway",
                    estimator=EstimatorOptions(shots=100),
                    sampler=SamplerOptions(shots=100),
                ),
            )
        )


async def test_client_result():
    hamiltonian = SparsePauliOp.from_list(
        [("YZ", 0.3980), ("ZI", -0.3980), ("ZZ", -0.0113), ("XX", 0.1810)]
    )
    with Client(Credentials.load(".credentials.json"), url=url) as client:
        result = VQE.builder().observables(hamiltonian).send(client).result()
        assert isinstance(result, VQEResult)



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Translator between an Ising Hamiltonian and a quadratic program"""

import math
from typing import Tuple

import numpy as np
from qiskit.quantum_info import Pauli, SparsePauliOp, Operator
from qiskit.quantum_info.operators.base_operator import BaseOperator

from ..exceptions import QiskitOptimizationError
from ..problems.quadratic_program import QuadraticProgram


def to_ising(quad_prog: QuadraticProgram) -> Tuple[SparsePauliOp, float]:
    """Return the Ising Hamiltonian of this problem.

    Variables are mapped to qubits in the same order, i.e.,
    i-th variable is mapped to i-th qubit.
    See https://github.com/Qiskit/qiskit-terra/issues/1148 for details.

    Args:
        quad_prog: The problem to be translated.

    Returns:
        A tuple (qubit_op, offset) comprising the qubit operator for the problem
        and offset for the constant value in the Ising Hamiltonian.

    Raises:
        QiskitOptimizationError: If an integer variable or a continuous variable exists
            in the problem.
        QiskitOptimizationError: If constraints exist in the problem.
    """
    # if problem has variables that are not binary, raise an error
    if quad_prog.get_num_vars() > quad_prog.get_num_binary_vars():
        raise QiskitOptimizationError(
            "The type of all variables must be binary. "
            "You can use `QuadraticProgramToQubo` converter "
            "to convert integer variables to binary variables. "
            "If the problem contains continuous variables, `to_ising` cannot handle it. "
            "You might be able to solve it with `ADMMOptimizer`."
        )

    # if constraints exist, raise an error
    if quad_prog.linear_constraints or quad_prog.quadratic_constraints:
        raise QiskitOptimizationError(
            "There must be no constraint in the problem. "
            "You can use `QuadraticProgramToQubo` converter "
            "to convert constraints to penalty terms of the objective function."
        )

    # initialize Hamiltonian.
    num_vars = quad_prog.get_num_vars()
    pauli_list = []
    offset = 0.0
    zero = np.zeros(num_vars, dtype=bool)

    # set a sign corresponding to a maximized or minimized problem.
    # sign == 1 is for minimized problem. sign == -1 is for maximized problem.
    sense = quad_prog.objective.sense.value

    # convert a constant part of the objective function into Hamiltonian.
    offset += quad_prog.objective.constant * sense

    # convert linear parts of the objective function into Hamiltonian.
    for idx, coef in quad_prog.objective.linear.to_dict().items():
        z_p = zero.copy()
        weight = coef * sense / 2
        z_p[idx] = True

        pauli_list.append(SparsePauliOp(Pauli((z_p, zero)), -weight))
        offset += weight

    # create Pauli terms
    for (i, j), coeff in quad_prog.objective.quadratic.to_dict().items():
        weight = coeff * sense / 4

        if i == j:
            offset += weight
        else:
            z_p = zero.copy()
            z_p[i] = True
            z_p[j] = True
            pauli_list.append(SparsePauliOp(Pauli((z_p, zero)), weight))

        z_p = zero.copy()
        z_p[i] = True
        pauli_list.append(SparsePauliOp(Pauli((z_p, zero)), -weight))

        z_p = zero.copy()
        z_p[j] = True
        pauli_list.append(SparsePauliOp(Pauli((z_p, zero)), -weight))

        offset += weight

    if pauli_list:
        # Remove paulis whose coefficients are zeros.
        qubit_op = sum(pauli_list).simplify(atol=0)
    else:
        # If there is no variable, we set num_nodes=1 so that qubit_op should be an operator.
        # If num_nodes=0, I^0 = 1 (int).
        num_vars = max(1, num_vars)
        qubit_op = SparsePauliOp("I" * num_vars, 0)

    return qubit_op, offset


def from_ising(
    qubit_op: BaseOperator,
    offset: float = 0.0,
    linear: bool = False,
) -> QuadraticProgram:
    r"""Create a quadratic program from a qubit operator and a shift value.

    Variables are mapped to qubits in the same order, i.e.,
    i-th variable is mapped to i-th qubit.
    See https://github.com/Qiskit/qiskit-terra/issues/1148 for details.

    Args:
        qubit_op: The qubit operator of the problem.
        offset: The constant term in the Ising Hamiltonian.
        linear: If linear is True, :math:`x^2` is treated as a linear term
            since :math:`x^2 = x` for :math:`x \in \{0,1\}`.
            Otherwise, :math:`x^2` is treat as a quadratic term.
            The default value is False.

    Returns:
        The quadratic program corresponding to the qubit operator.

    Raises:
        QiskitOptimizationError: if there are Pauli Xs or Ys in any Pauli term
        QiskitOptimizationError: if there are more than 2 Pauli Zs in any Pauli term
        QiskitOptimizationError: if any Pauli term has an imaginary coefficient
    """
    # quantum_info
    if isinstance(qubit_op, BaseOperator):
        if not isinstance(qubit_op, SparsePauliOp):
            qubit_op = SparsePauliOp.from_operator(Operator(qubit_op))

    quad_prog = QuadraticProgram()
    quad_prog.binary_var_list(qubit_op.num_qubits)

    # prepare a matrix of coefficients of Pauli terms
    # `pauli_coeffs_diag` is the diagonal part
    # `pauli_coeffs_triu` is the upper triangular part
    pauli_coeffs_diag = [0.0] * qubit_op.num_qubits
    pauli_coeffs_triu = {}

    for pauli_op in qubit_op:
        pauli = pauli_op.paulis[0]
        coeff = pauli_op.coeffs[0]

        if not math.isclose(coeff.imag, 0.0, abs_tol=1e-10):
            raise QiskitOptimizationError(f"Imaginary coefficient exists: {pauli_op}")

        if np.any(pauli.x):
            raise QiskitOptimizationError(
                f"Pauli X or Y exists in the Pauli term: {pauli}"
            )

        # indices of Pauli Zs in the Pauli term
        z_index = np.where(pauli.z)[0]
        num_z = len(z_index)

        if num_z == 0:
            offset += coeff.real
        elif num_z == 1:
            pauli_coeffs_diag[z_index[0]] = coeff.real
        elif num_z == 2:
            pauli_coeffs_triu[z_index[0], z_index[1]] = coeff.real
        else:
            raise QiskitOptimizationError(
                f"There are more than 2 Pauli Zs in the Pauli term: {pauli}"
            )

    linear_terms = {}
    quadratic_terms = {}

    # For quadratic pauli terms of operator
    # x_i * x_j = (1 - Z_i - Z_j + Z_i * Z_j)/4
    for (i, j), weight in pauli_coeffs_triu.items():
        # Add a quadratic term to the objective function of `QuadraticProgram`
        # The coefficient of the quadratic term in `QuadraticProgram` is
        # 4 * weight of the pauli
        quadratic_terms[i, j] = 4 * weight
        pauli_coeffs_diag[i] += weight
        pauli_coeffs_diag[j] += weight
        offset -= weight

    # After processing quadratic pauli terms, only linear paulis are left
    # x_i = (1 - Z_i)/2
    for i, weight in enumerate(pauli_coeffs_diag):
        # Add a linear term to the objective function of `QuadraticProgram`
        # The coefficient of the linear term in `QuadraticProgram` is
        # 2 * weight of the pauli
        if linear:
            linear_terms[i] = -2 * weight
        else:
            quadratic_terms[i, i] = -2 * weight
        offset += weight

    quad_prog.minimize(constant=offset, linear=linear_terms, quadratic=quadratic_terms)

    return quad_prog



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2020, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""An abstract class for optimization algorithms in Qiskit optimization module."""

from abc import ABC, abstractmethod
from typing import List, Union

import numpy as np

from ..problems.quadratic_program import QuadraticProgram


class QuadraticProgramConverter(ABC):
    """
    An abstract class for converters of quadratic programs in Qiskit optimization module.
    """

    @abstractmethod
    def convert(self, problem: QuadraticProgram) -> QuadraticProgram:
        """Convert a QuadraticProgram into another form
        and keep the information required to interpret the result.
        """
        raise NotImplementedError

    @abstractmethod
    def interpret(self, x: Union[np.ndarray, List[float]]) -> np.ndarray:
        """Interpret a result into another form using the information of conversion"""
        raise NotImplementedError



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2021, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Quadratic program translators (:mod:`qiskit_optimization.translators`)
======================================================================

.. currentmodule:: qiskit_optimization.translators

Translators between :class:`~qiskit_optimization.problems.QuadraticProgram` and
other optimization models or other objects.

Translators
----------------------
.. autosummary::
   :toctree: ../stubs/
   :nosignatures:

   from_ising
   to_ising
"""

from .ising import from_ising, to_ising

__all__ = [
    "from_ising",
    "to_ising",
]



from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Any, List, Tuple, Dict

from pydantic import BaseModel

import q3as.algo.vqe

from q3as.encoding.numpy import EncodedArray
from q3as.encoding.qiskit import (
    EncodedEstimatorResult,
    EncodedSamplerResult,
    EncodedObservables,
    EncodedQuantumCircuit,
)

from .app import EncodedApplication
from .optimizer import EncodedOptimizer
from q3as.algo.types import HaltReason

if TYPE_CHECKING:
    from q3as.algo.vqe import VQE, VQEResult, VQEIteration


class EncodedVQEIteration(BaseModel):
    iter: int
    cost: float
    params: EncodedArray
    estimated: EncodedEstimatorResult
    best: bool

    @classmethod
    def encode(cls, result: VQEIteration) -> EncodedVQEIteration:
        return cls(
            iter=result.iter,
            cost=result.cost,
            params=EncodedArray.encode(result.params),
            estimated=EncodedEstimatorResult.encode(result.estimated),
            best=result.best,
        )

    def decode(self) -> VQEIteration:
        return q3as.algo.vqe.VQEIteration(
            iter=self.iter,
            cost=self.cost,
            params=self.params.decode(),
            estimated=self.estimated.decode(),
            best=self.best,
        )


class EncodedVQEResult(BaseModel):
    params: EncodedArray
    iter: int
    reason: HaltReason
    cost: Optional[float]
    estimated: Optional[EncodedEstimatorResult]
    sampled: Optional[EncodedSamplerResult]
    meas_counts: Optional[Dict[str, float]]
    interpreted: Optional[List[Tuple[Any, int]]]

    @classmethod
    def encode(cls, result: VQEResult) -> EncodedVQEResult:
        return cls(
            params=EncodedArray.encode(result.params),
            iter=result.iter,
            reason=result.reason,
            cost=result.cost,
            estimated=None
            if result.estimated is None
            else EncodedEstimatorResult.encode(result.estimated),
            sampled=None
            if result.sampled is None
            else EncodedSamplerResult.encode(result.sampled),
            meas_counts=result.meas_counts,
            interpreted=result.interpreted,
        )

    def decode(self):
        return q3as.algo.vqe.VQEResult(
            iter=self.iter,
            cost=self.cost,
            params=self.params.decode(),
            reason=self.reason,
            estimated=None if self.estimated is None else self.estimated.decode(),
            sampled=None if self.sampled is None else self.sampled.decode(),
            meas_counts=self.meas_counts,
            interpreted=self.interpreted,
        )


class EncodedVQE(BaseModel):
    ansatz: EncodedQuantumCircuit
    observables: EncodedObservables
    initial_params: EncodedArray
    optimizer: EncodedOptimizer
    maxiter: int
    app: Optional[EncodedApplication]

    @classmethod
    def encode(cls, vqe: VQE) -> EncodedVQE:
        return cls(
            ansatz=EncodedQuantumCircuit.encode(vqe.ansatz),
            observables=EncodedObservables.encode(vqe.observables),
            initial_params=EncodedArray.encode(vqe.initial_params),
            optimizer=EncodedOptimizer.encode(vqe.optimizer),
            maxiter=vqe.maxiter,
            app=None if vqe.app is None else EncodedApplication.encode(vqe.app),
        )

    def decode(self) -> VQE:
        return q3as.algo.vqe.VQE(
            ansatz=self.ansatz.decode(),
            observables=self.observables.decode(),
            initial_params=self.initial_params.decode(),
            optimizer=self.optimizer,
            maxiter=self.maxiter,
            app=None if self.app is None else self.app.decode(),
        )



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2019, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Linear expression interface."""

from typing import List, Union, Dict, Any
from dataclasses import dataclass

from numpy import ndarray
from scipy.sparse import spmatrix, dok_matrix

from .quadratic_program_element import QuadraticProgramElement
from ..exceptions import QiskitOptimizationError
from ..infinity import INFINITY


@dataclass
class ExpressionBounds:
    """Lower bound and upper bound of a linear expression or a quadratic expression"""

    lowerbound: float
    """Lower bound"""

    upperbound: float
    """Upper bound"""


class LinearExpression(QuadraticProgramElement):
    """Representation of a linear expression by its coefficients."""

    def __init__(
        self,
        quadratic_program: Any,
        coefficients: Union[
            ndarray, spmatrix, List[float], Dict[Union[int, str], float]
        ],
    ) -> None:
        """Creates a new linear expression.

        The linear expression can be defined via an array, a list, a sparse matrix, or a dictionary
        that uses variable names or indices as keys and stores the values internally as a
        dok_matrix.

        Args:
            quadratic_program: The parent QuadraticProgram.
            coefficients: The (sparse) representation of the coefficients.

        """
        super().__init__(quadratic_program)
        self.coefficients = coefficients

    def __getitem__(self, i: Union[int, str]) -> float:
        """Returns the i-th coefficient where i can be a variable name or index.

        Args:
            i: the index or name of the variable corresponding to the coefficient.

        Returns:
            The coefficient corresponding to the addressed variable.
        """
        if isinstance(i, str):
            i = self.quadratic_program.variables_index[i]
        return self.coefficients[0, i]

    def __setitem__(self, i: Union[int, str], value: float) -> None:
        if isinstance(i, str):
            i = self.quadratic_program.variables_index[i]
        self._coefficients[0, i] = value

    def _coeffs_to_dok_matrix(
        self, coefficients: Union[ndarray, spmatrix, List, Dict[Union[int, str], float]]
    ) -> dok_matrix:
        """Maps given 1d-coefficients to a dok_matrix.

        Args:
            coefficients: The 1d-coefficients to be mapped.

        Returns:
            The given 1d-coefficients as a dok_matrix

        Raises:
            QiskitOptimizationError: if coefficients are given in unsupported format.
        """
        if (
            isinstance(coefficients, list)
            or isinstance(coefficients, ndarray)
            and len(coefficients.shape) == 1
        ):
            coefficients = dok_matrix([coefficients])
        elif isinstance(coefficients, spmatrix):
            coefficients = dok_matrix(coefficients)
        elif isinstance(coefficients, dict):
            coeffs = dok_matrix((1, self.quadratic_program.get_num_vars()))
            for index, value in coefficients.items():
                if isinstance(index, str):
                    index = self.quadratic_program.variables_index[index]
                coeffs[0, index] = value
            coefficients = coeffs
        else:
            raise QiskitOptimizationError("Unsupported format for coefficients.")
        return coefficients

    @property
    def coefficients(self) -> dok_matrix:
        """Returns the coefficients of the linear expression.

        Returns:
            The coefficients of the linear expression.
        """
        return self._coefficients

    @coefficients.setter
    def coefficients(
        self,
        coefficients: Union[
            ndarray, spmatrix, List[float], Dict[Union[str, int], float]
        ],
    ) -> None:
        """Sets the coefficients of the linear expression.

        Args:
            coefficients: The coefficients of the linear expression.
        """
        self._coefficients = self._coeffs_to_dok_matrix(coefficients)

    def to_array(self) -> ndarray:
        """Returns the coefficients of the linear expression as array.

        Returns:
            An array with the coefficients corresponding to the linear expression.
        """
        return self._coefficients.toarray()[0]

    def to_dict(self, use_name: bool = False) -> Dict[Union[int, str], float]:
        """Returns the coefficients of the linear expression as dictionary, either using variable
        names or indices as keys.

        Args:
            use_name: Determines whether to use index or names to refer to variables.

        Returns:
            An dictionary with the coefficients corresponding to the linear expression.
        """
        if use_name:
            return {
                self.quadratic_program.variables[k].name: v
                for (_, k), v in self._coefficients.items()
            }
        else:
            return {k: v for (_, k), v in self._coefficients.items()}

    def evaluate(self, x: Union[ndarray, List, Dict[Union[int, str], float]]) -> float:
        """Evaluate the linear expression for given variables.

        Args:
            x: The values of the variables to be evaluated.

        Returns:
            The value of the linear expression given the variable values.
        """
        # cast input to dok_matrix if it is a dictionary
        x = self._coeffs_to_dok_matrix(x)

        # compute the dot-product of the input and the linear coefficients
        val = (x @ self.coefficients.transpose())[0, 0]

        # return the result
        return val

    # pylint: disable=unused-argument
    def evaluate_gradient(
        self, x: Union[ndarray, List, Dict[Union[int, str], float]]
    ) -> ndarray:
        """Evaluate the gradient of the linear expression for given variables.

        Args:
            x: The values of the variables to be evaluated.

        Returns:
            The value of the gradient of the linear expression given the variable values.
        """

        # extract the coefficients as array and return it
        return self.to_array()

    @property
    def bounds(self) -> ExpressionBounds:
        """Returns the lower bound and the upper bound of the linear expression

        Returns:
            The lower bound and the upper bound of the linear expression

        Raises:
            QiskitOptimizationError: if the linear expression contains any unbounded variable

        """
        l_b = u_b = 0.0
        for ind, coeff in self.to_dict().items():
            x = self.quadratic_program.get_variable(ind)
            if x.lowerbound == -INFINITY or x.upperbound == INFINITY:
                raise QiskitOptimizationError(
                    f"Linear expression contains an unbounded variable: {x.name}"
                )
            lst = [coeff * x.lowerbound, coeff * x.upperbound]
            l_b += min(lst)
            u_b += max(lst)
        return ExpressionBounds(lowerbound=l_b, upperbound=u_b)

    def __repr__(self):
        # pylint: disable=cyclic-import
        from ..translators.prettyprint import expr2str, DEFAULT_TRUNCATE

        return f"<{self.__class__.__name__}: {expr2str(linear=self, truncate=DEFAULT_TRUNCATE)}>"

    def __str__(self):
        # pylint: disable=cyclic-import
        from ..translators.prettyprint import expr2str

        return f"{expr2str(linear=self)}"



from typing import Literal
from pydantic import BaseModel


class SamplerOptions(BaseModel):
    """
    Options for the sampler.
    """

    shots: int = 1024
    "Number of shots for the sampler."


class EstimatorOptions(BaseModel):
    """
    Options for the estimator.
    """

    shots: int = 1024
    "Number of shots for the estimator."


type BackendName = Literal[
    "auto",
    "q3as",
    "q3as:aer_simulator",
    "q3as:aer_simulator_statevector",
    "q3as:aer_simulator_density_matrix",
    "q3as:aer_simulator_stabilizer",
    "q3as:aer_simulator_matrix_product_state",
    "q3as:aer_simulator_extended_stabilizer",
    "q3as:aer_simulator_unitary",
    "q3as:aer_simulator_superop",
    "qasm_simulator",
    "statevector_simulator",
    "unitary_simulator",
    "scaleway",
    "scaleway:aer_simulation_2l4",
    "scaleway:aer_simulation_2l40s",
    "scaleway:aer_simulation_4l40s",
    "scaleway:aer_simulation_8l40s",
    "scaleway:aer_simulation_h100",
    "scaleway:aer_simulation_2h100",
    "scaleway:aer_simulation_pop_c16m128",
    "scaleway:aer_simulation_pop_c32m256",
    "scaleway:aer_simulation_pop_c64m512",
]


class RunOptions(BaseModel):
    """
    Options for  running in the cloud.
    """

    backend: BackendName = "auto"
    "Name of the backend to use. See `q3as.run_options.BackendName` for possible values."
    sampler: SamplerOptions = SamplerOptions()
    "Options for the sampler."
    estimator: EstimatorOptions = EstimatorOptions()
    "Options for the estimator."



# This code is part of a Qiskit project.
#
# (C) Copyright IBM 2021, 2023.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""Converters to flip problem sense, e.g. maximization to minimization and vice versa."""

import copy
from typing import Optional, List, Union

import numpy as np

from .quadratic_program_converter import QuadraticProgramConverter
from ..exceptions import QiskitOptimizationError
from ..problems.quadratic_objective import ObjSense
from ..problems.quadratic_program import QuadraticProgram


class _FlipProblemSense(QuadraticProgramConverter):
    """Flip the sense of a problem, e.g. converts from maximization to minimization and
    vice versa, regardless of the current sense."""

    def __init__(self) -> None:
        self._src_num_vars: Optional[int] = None

    def convert(self, problem: QuadraticProgram) -> QuadraticProgram:
        """Flip the sense of a problem.

        Args:
            problem: The problem to be flipped.

        Returns:
            A converted problem, that has the flipped sense.
        """

        # copy original number of variables as reference.
        self._src_num_vars = problem.get_num_vars()
        desired_sense = self._get_desired_sense(problem)

        # flip the problem sense
        if problem.objective.sense != desired_sense:
            desired_problem = copy.deepcopy(problem)
            desired_problem.objective.sense = desired_sense
            desired_problem.objective.constant = (-1) * problem.objective.constant
            desired_problem.objective.linear = (
                -1
            ) * problem.objective.linear.coefficients
            desired_problem.objective.quadratic = (
                -1
            ) * problem.objective.quadratic.coefficients
        else:
            desired_problem = problem

        return desired_problem

    def _get_desired_sense(self, problem: QuadraticProgram) -> ObjSense:
        """
        Computes a desired sense of the problem. By default, flip the sense.

        Args:
            problem: a problem to check

        Returns:
            A desired sense, if the problem was a minimization problem, then the sense is
            maximization and vice versa.
        """
        if problem.objective.sense == ObjSense.MAXIMIZE:
            return ObjSense.MINIMIZE
        else:
            return ObjSense.MAXIMIZE

    def interpret(self, x: Union[np.ndarray, List[float]]) -> np.ndarray:
        """Convert the result of the converted problem back to that of the original problem.

        Args:
            x: The result of the converted problem or the given result in case of FAILURE.

        Returns:
            The result of the original problem.

        Raises:
            QiskitOptimizationError: if the number of variables in the result differs from
                                     that of the original problem.
        """
        if len(x) != self._src_num_vars:
            raise QiskitOptimizationError(
                f"The number of variables in the passed result differs from "
                f"that of the original problem, should be {self._src_num_vars}, but got {len(x)}."
            )
        return np.asarray(x)


class MaximizeToMinimize(_FlipProblemSense):
    """Convert a maximization problem to a minimization problem only if it is a maximization
    problem, otherwise problem's sense is unchanged."""

    def _get_desired_sense(self, problem: QuadraticProgram) -> ObjSense:
        return ObjSense.MINIMIZE


class MinimizeToMaximize(_FlipProblemSense):
    """Convert a minimization problem to a maximization problem only if it is a minimization
    problem, otherwise problem's sense is unchanged."""

    def _get_desired_sense(self, problem: QuadraticProgram) -> ObjSense:
        return ObjSense.MAXIMIZE



from __future__ import annotations
from typing import TypeVar, Generic, Literal, List, Tuple, Any
from abc import ABC, abstractmethod

import numpy as np

from qiskit.primitives import BitArray
from qiskit.primitives.containers.observables_array import (
    ObservablesArrayLike,
)

type ApplicationName = Literal["qubo", "maxcut"]


class BitString:
    data: str

    def __init__(self, data: str) -> None:
        self.data = data

    def to_list(self) -> np.ndarray:
        return np.array([b == "1" for b in self.data], dtype=bool)


Encoded = TypeVar("Encoded")
State = TypeVar("State")


class Application(ABC, Generic[Encoded, State]):
    @abstractmethod
    def name(self) -> ApplicationName:
        pass

    @abstractmethod
    def encode(self) -> Encoded:
        pass

    @classmethod
    @abstractmethod
    def decode(cls, encoded: Encoded) -> Application:
        pass

    @classmethod
    def validate_encoded(cls, encoded: Any) -> Encoded:
        return encoded

    @classmethod
    def decode_any(cls, encoded: Any) -> Application:
        return cls.decode(cls.validate_encoded(encoded))

    @abstractmethod
    def hamiltonian(self) -> ObservablesArrayLike:
        pass

    @abstractmethod
    def interpret(self, bit_string: BitString) -> State:
        pass

    def interpreted_meas(self, meas: BitArray) -> List[Tuple[State, int]]:
        out = {}
        for state, count in meas.get_counts().items():
            interpreted = self.interpret(BitString(state))
            cur_count = out.get(interpreted, 0)
            out[interpreted] = cur_count + count
        return list(out.items())
